#include "Converter.hpp"
#include "utils.hpp"

#include <openbabel/mol.h>
#include <openbabel/bond.h>
#include <openbabel/obconversion.h>

#include <algorithm>
#include <string>
#include <cctype>
#include <cstdio>

namespace {
  using namespace std;

  vector<fragdock::Atom> getFragmentMolAtoms(vector<format::PDBQTAtom>& patoms) {
    vector<fragdock::Atom> fatoms;
    for(int i = 0; i < patoms.size(); i++){
      format::PDBQTAtom& patom = patoms[i];
      std::string atomtype_name = patom.element;
      /*
	std::string grid_types[] = {"A", "C", "HD", "N", "NA", "OA", "P", "SA", "S", "F", "Cl", "DU"};
	if(std::find(&grid_types[0], &grid_types[12], patom.element) == &grid_types[12]){ //if it cannot be found
	atomtype_name = "DU";
	}
      */
      fragdock::AtomType type(atomtype_name);
      fragdock::Vector3d pos(patom.x, patom.y, patom.z);
      fatoms.push_back( fragdock::Atom(patom.id, pos, type, patom.charge) );
    }
    return fatoms;
  }
      
  std::string getAtomTypeName(const OpenBabel::OBAtom& oatom) {
    OpenBabel::OBAtom& atom = const_cast<OpenBabel::OBAtom&>(oatom);

    OpenBabel::ttab.SetFromType("INT");
    OpenBabel::ttab.SetToType("XYZ");
    
    if     (atom.IsCarbon()   &&  atom.IsAromatic())      return "A";
    else if(atom.IsCarbon()   && !atom.IsAromatic())      return "C";
    else if(atom.IsNitrogen() &&  atom.IsHbondAcceptor()) return "NA";
    else if(atom.IsNitrogen() && !atom.IsHbondAcceptor()) return "N";
    else if(atom.IsOxygen()   &&  atom.IsHbondAcceptor()) return "OA";
    else if(atom.IsOxygen()   && !atom.IsHbondAcceptor()) return "O";
    else if(atom.IsSulfur()   &&  atom.IsHbondAcceptor()) return "SA";
    else if(atom.IsSulfur()   && !atom.IsHbondAcceptor()) return "S";
    else if(atom.IsHydrogen() &&  atom.IsHbondDonorH())   return "HD";
    else if(atom.IsHydrogen() && !atom.IsHbondDonorH())   return "H";
    else {
      std::string src = atom.GetType();
      std::string dst;
      OpenBabel::ttab.Translate(dst, src);
      return dst;
    }
  }
  
}

namespace format {

  fragdock::Molecule Converter::toFragmentMol(const format::PDBQTMolecule &pdbqt){
    std::vector<format::PDBQTAtom> patoms = pdbqt.atoms;
    std::vector<fragdock::Atom> atoms = getFragmentMolAtoms(patoms);
    fragdock::Molecule mol(atoms);

    std::vector<format::PDBQTBond> pbonds = pdbqt.bonds;
    for(int i = 0; i < pbonds.size(); i++){
      format::PDBQTBond bond = pbonds.at(i);
      mol.append(fragdock::Bond(bond.origin_atom_index,
				bond.target_atom_index,
				true ));
    }

    return mol;
  }
  
  fragdock::Molecule Converter::toFragmentMol(const OpenBabel::OBMol &c_obmol){
    std::vector<fragdock::Atom> fatoms;
    OpenBabel::OBMol &obmol = const_cast<OpenBabel::OBMol&>(c_obmol);


    for(int i=0; i<obmol.NumAtoms(); i++) {
      OpenBabel::OBAtom& oatom = *( obmol.GetAtom(i+1) ); //obmol::index is 1 origin

      std::string atomtype = getAtomTypeName(oatom);
      fragdock::AtomType type(atomtype);
      fragdock::Vector3d pos(oatom.GetX(), oatom.GetY(), oatom.GetZ());
      fatoms.push_back( fragdock::Atom(oatom.GetId(), pos, type, oatom.GetPartialCharge()) );
    }
    fragdock::Molecule mol(fatoms);
    for(int i = 0; i < obmol.NumBonds(); i++){
      mol.append(fragdock::Bond(obmol.GetBond(i)->GetBeginAtom()->GetId(),
				obmol.GetBond(i)->GetEndAtom()->GetId(),
				obmol.GetBond(i)->IsRotor() ));
    }

    return mol;
  }

  format::PDBQTMolecule Converter::toPDBQT(const fragdock::Molecule &mol, const format::PDBQTMolecule& original_pdbqt) {
    using namespace fragdock;
    using namespace std;
    vector<Atom> atoms = mol.getAtoms();
    format::PDBQTMolecule pdbqt = original_pdbqt;
    vector<format::PDBQTAtom> new_atoms;
    vector<format::PDBQTAtom> old_atoms = pdbqt.atoms;
    
    for(int i = 0; i < old_atoms.size(); i++) {
      format::PDBQTAtom natom = old_atoms[i];
      natom.x = atoms[i].x;
      natom.y = atoms[i].y;
      natom.z = atoms[i].z;
      new_atoms.push_back(natom);
    }

    pdbqt.atoms = new_atoms;

    return pdbqt;
  }

  OpenBabel::OBMol Converter::toOBMol(const fragdock::Molecule &mol, const OpenBabel::OBMol& original_obmol) {
    using namespace fragdock;
    using namespace std;

    OpenBabel::OBMol obmol = original_obmol;
    obmol.SetTitle("");

    std::vector<int> del_id_list;
    for(int i=0; i<original_obmol.NumAtoms(); i++){
      del_id_list.push_back(i);
    }


    for(int i=0; i<mol.getAtoms().size(); i++) {
      int id = mol.getAtoms()[i].getId();

      del_id_list.erase(std::remove(del_id_list.begin(), del_id_list.end(), id), del_id_list.end()); 
      /*
      std::cout << mol.getAtoms()[i].getType().getName() << " " 
		<< original_obmol.GetAtomById(id)->GetType() << " "
		<< std::endl;
      */
    }

    for(int i=0; i<del_id_list.size(); i++){
      int id = del_id_list[i];
      //std::cout << "del list" << i << ": " << id << " : " << obmol.GetAtomById(id)->GetType() << std::endl;
      obmol.DeleteAtom(obmol.GetAtomById(id));
    }

    /*
    vector<Atom> atoms = mol.getAtoms();
    for(int i = 0; i < obmol.NumAtoms(); i++) {
      OpenBabel::OBAtom* atom = obmol.GetAtomById(i);
      double x = atoms[i].x;
      double y = atoms[i].y;
      double z = atoms[i].z;
      atom->SetVector(x, y, z);
    }
    */


    return obmol;
  }

}
