#include "MoleculeToFragments.hpp"
#include "UnionFindTree.hpp"
namespace {
  using namespace std;
  using namespace fragdock;

  vector<vector<int> > edges;
  vector<int> ring;
  vector<int> route;
  vector<bool> done;
  
  void dfs(int now, int start) {
    if(done[now]) {
      if(now == start and route.size() > 2) {
	ring = route;
      }
      return;
    }
    done[now] = true;
    
    route.push_back(now);
    for(int i = 0; i < edges[now].size(); i++)
      dfs(edges[now][i], start);
    
    route.pop_back();
  }
  
  vector<vector<int> > ringDetector(int atom_num, const vector<Bond> &bonds) {
    edges = vector<vector<int> >(atom_num, vector<int>());
    for(int i = 0; i < bonds.size(); i++) {
      const Bond &bond = bonds[i];
      int a = bond.atom_id1;
      int b = bond.atom_id2;
      edges[a].push_back(b);
      edges[b].push_back(a);
    }
    
    vector<vector<int> > ret;
    for(int i = 0; i < atom_num; i++) {
      done = vector<bool>(atom_num, false);
      ring = vector<int>();
      route = vector<int>();
      dfs(i, i);
      if(ring.size() > 0) ret.push_back(ring);
    }
    
    return ret;
  }

  
}

namespace fragdock {
  vector<Fragment> getFragments(const Molecule &mol) {
    const vector<Atom> &atoms = mol.getAtoms();
    const vector<Bond> &bonds = mol.getBonds();
    // unite two atoms if the bond between them is NOT 'Single'
    utils::UnionFindTree uf((int)atoms.size());
    for(int i = 0; i < bonds.size(); i++) {
      const Bond &bond = bonds[i];
      int a = bond.atom_id1;
      int b = bond.atom_id2;
      if(!bond.is_rotor)
	uf.unite(a, b);      
    }
    // unite members of ring systems
    vector<vector<int> > rings = ringDetector(atoms.size(), bonds);    
    for(int i = 0; i < rings.size(); i++)
      for(int j = 0; j < rings[i].size(); j++)
	uf.unite(rings[i][0], rings[i][j]);
    // count adjacent H
    vector<int> h_cnt(atoms.size(), false);
    for(int i = 0; i < bonds.size(); i++) {
      const Bond &bond = bonds[i];
      int a = bond.atom_id1;
      int b = bond.atom_id2;
      if(atoms[a].getType().getName() != "H")
	h_cnt[b]++;
      if(atoms[b].getType().getName() != "H")
	h_cnt[a]++;
    }
    // unite one atom except H
    vector<bool> done(atoms.size(), false);
    for(int i = 0; i < bonds.size(); i++) {
      const Bond &bond = bonds[i];
      int a = bond.atom_id1;
      int b = bond.atom_id2;
      if(atoms[a].getType().getName() != "H" and
	 atoms[b].getType().getName() != "H")
	for(int _ = 0; _ < 2; _++) {
	  if((uf.getSize(a) == 1 and h_cnt[a] <= 2 and
	     uf.getSize(b) > 1 and !done[b]) or
	     (uf.getSize(a) == 1 and uf.getSize(b) == 1 and
	      h_cnt[a] <= 2 and h_cnt[b] <= 2)){
	    uf.unite(a, b);
	    done[a] = true;
	  }
	  swap(a, b);
	}
    }
    // unite H
    for(int i = 0; i < bonds.size(); i++) {
      const Bond &bond = bonds[i];
      int a = bond.atom_id1;
      int b = bond.atom_id2;
      if(atoms[a].getType().getName() == "H" or
	 atoms[b].getType().getName() == "H") uf.unite(a, b);
    }
    // get fragments
    vector<vector<int> > id_sets = uf.getSets();
    id_sets = uf.getSets();
    // atom_id to set_id
    vector<int> set_id(atoms.size());
    for(int i = 0; i < id_sets.size(); i++)
      for(int j = 0; j < id_sets[i].size(); j++)
	set_id[id_sets[i][j]] = i;

    // signle bond count (to find rotatable edge)
    vector<vector<Vector3d> > axises(atoms.size());
    for(int i = 0; i < bonds.size(); i++) {
      const Bond &bond = bonds[i];
      int a = bond.atom_id1;
      int b = bond.atom_id2;
      if(bond.is_rotor and
	 atoms[a].getType().getName() != "H" and
	 atoms[b].getType().getName() != "H" and
	 set_id[a] == set_id[b]) {
	Vector3d vec_a = atoms[b] - atoms[a];
	Vector3d vec_b = atoms[a] - atoms[b];
	vec_a /= vec_a.abs();
	vec_b /= vec_b.abs();
	axises[a].push_back(vec_a);
	axises[b].push_back(vec_b);
      }
    }
    // create edges
    vector<vector<FragmentEdge> > edgess(id_sets.size());
    vector<vector<int> > near_ids(id_sets.size());
    for(int i = 0; i < bonds.size(); i++) {
      const Bond &bond = bonds[i];
      int a = bond.atom_id1;
      int b = bond.atom_id2;
      if(set_id[a] != set_id[b]) {
	Vector3d vec_a = atoms[a];
	Vector3d vec_b = atoms[b];
	FragmentEdge edge_a(set_id[a], set_id[b], edgess[set_id[b]].size(), vec_a, vec_b);
	FragmentEdge edge_b(set_id[b], set_id[a], edgess[set_id[a]].size(), vec_b, vec_a);
	if(axises[a].size() == 1) edge_a.setAxis(axises[a][0]);
	if(axises[b].size() == 1) edge_b.setAxis(axises[b][0]);
	edgess[set_id[a]].push_back(edge_a);
	edgess[set_id[b]].push_back(edge_b);
	near_ids[set_id[a]].push_back(b);
	near_ids[set_id[b]].push_back(a);
      }
    }
    // create fragments
    vector<Fragment> fragments;
    for(int i = 0; i < id_sets.size(); i++) {
      vector<Atom> frag_atoms;
      for(int j = 0; j < id_sets[i].size(); j++)
	frag_atoms.push_back(atoms[id_sets[i][j]]);
      Fragment frag(i, edgess[i], frag_atoms, near_ids[i]);
      fragments.push_back(frag);
    }
    
    return fragments;
  }
}
