#include "common.hpp"
#include "energy_grid.hpp"
#include "utils.hpp"

#include <iostream>
#include <fstream>
#include <string>
#include <cstdlib>
#include <cstdio>
#include <cmath>

#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp> // is_any_of

namespace docking{
 
  docking::EnergyGrid ReadEnergyGridFile(const char *filename){
    using namespace std;
    docking::EnergyGrid retval;

    std::ifstream ifs(filename, std::ios::binary);
    ifs.exceptions(std::ifstream::failbit);
    //TODO Even if there isn't GridFile, program should still run and is_valid flag should be false;

    string buffer;

    ifs.read(reinterpret_cast<char*>(&retval.center), sizeof(retval.center));
    ifs.read(reinterpret_cast<char*>(&retval.pitch ), sizeof(retval.pitch ));
    ifs.read(reinterpret_cast<char*>(&retval.num   ), sizeof(retval.num   ));
    
    int total_num = retval.num.x * retval.num.y * retval.num.z;
    retval.energy_list = new fltype[total_num];
    ifs.read(reinterpret_cast<char*>(retval.energy_list), sizeof(fltype)*total_num);

    retval.is_valid = true;
    return retval;
  }
  
  void WriteEnergyGridFile(const char *filename, const docking::EnergyGrid *grid){
    std::ofstream ofs(filename, std::ios::binary);
    if(ofs.fail()){
      std::cerr << "Error: filed to open new Grid file:" << filename << std::endl;
      abort();
    }
    
    ofs.write(reinterpret_cast<const char*>(&grid->center), sizeof(grid->center));
    ofs.write(reinterpret_cast<const char*>(&grid->pitch),  sizeof(grid->pitch ));
    ofs.write(reinterpret_cast<const char*>(&grid->num),    sizeof(grid->num   ));
    ofs.write(reinterpret_cast<const char*>(grid->energy_list), sizeof(fltype)*grid->num.x*grid->num.y*grid->num.z);

    ofs.close();
  }
  
  fltype GetValueNearest(const EnergyGrid *grid, const utils::Vec3D<fltype> *point){
    int suf_x = utils::round((point->x - grid->center.x)/grid->pitch.x + (grid->num.x)/2);
    int suf_y = utils::round((point->y - grid->center.y)/grid->pitch.y + (grid->num.y)/2);
    int suf_z = utils::round((point->z - grid->center.z)/grid->pitch.z + (grid->num.z)/2);

    suf_x = suf_x < 0 ? 0 : suf_x >= grid->num.x ? grid->num.x-1 : suf_x;
    suf_y = suf_y < 0 ? 0 : suf_y >= grid->num.y ? grid->num.y-1 : suf_y;
    suf_z = suf_z < 0 ? 0 : suf_z >= grid->num.z ? grid->num.z-1 : suf_z;

    return grid->energy_list[suf_z + grid->num.z*suf_y + grid->num.z*grid->num.y*suf_x];
  }
  fltype GetValueTrilinear(const EnergyGrid *grid, const utils::Vec3D<fltype> *point){
    float rsuf_x = (point->x - grid->center.x)/grid->pitch.x + (grid->num.x)/2;
    float rsuf_y = (point->y - grid->center.y)/grid->pitch.y + (grid->num.y)/2;
    float rsuf_z = (point->z - grid->center.z)/grid->pitch.z + (grid->num.z)/2;

    // left-upper side suffix
    int suf_x = static_cast<int>(rsuf_x);
    int suf_y = static_cast<int>(rsuf_y);
    int suf_z = static_cast<int>(rsuf_z);
    suf_x = suf_x < 0 ? 0 : suf_x >= grid->num.x-1 ? grid->num.x-2 : suf_x;
    suf_y = suf_y < 0 ? 0 : suf_y >= grid->num.y-1 ? grid->num.y-2 : suf_y;
    suf_z = suf_z < 0 ? 0 : suf_z >= grid->num.z-1 ? grid->num.z-2 : suf_z;

    fltype frac_x = rsuf_x - suf_x;
    fltype frac_y = rsuf_y - suf_y;
    fltype frac_z = rsuf_z - suf_z;
    frac_x = frac_x < 0 ? 0 : frac_x >= 1 ? 1 : frac_x;
    frac_y = frac_y < 0 ? 0 : frac_y >= 1 ? 1 : frac_y;
    frac_z = frac_z < 0 ? 0 : frac_z >= 1 ? 1 : frac_z;
      
      
    fltype energy = 0;
    energy += grid->energy_list[ suf_z    + grid->num.z* suf_y    + grid->num.z*grid->num.y* suf_x   ] * (1-frac_x) * (1-frac_y) * (1-frac_z);
    energy += grid->energy_list[(suf_z+1) + grid->num.z* suf_y    + grid->num.z*grid->num.y* suf_x   ] * (1-frac_x) * (1-frac_y) * frac_z;
    energy += grid->energy_list[ suf_z    + grid->num.z*(suf_y+1) + grid->num.z*grid->num.y* suf_x   ] * (1-frac_x) * frac_y     * (1-frac_z);
    energy += grid->energy_list[ suf_z    + grid->num.z* suf_y    + grid->num.z*grid->num.y*(suf_x+1)] * frac_x     * (1-frac_y) * (1-frac_z);
    energy += grid->energy_list[(suf_z+1) + grid->num.z*(suf_y+1) + grid->num.z*grid->num.y* suf_x   ] * (1-frac_x) * frac_y     * frac_z;
    energy += grid->energy_list[ suf_z    + grid->num.z*(suf_y+1) + grid->num.z*grid->num.y*(suf_x+1)] * frac_x     * frac_y     * (1-frac_z);
    energy += grid->energy_list[(suf_z+1) + grid->num.z* suf_y    + grid->num.z*grid->num.y*(suf_x+1)] * frac_x     * (1-frac_y) * frac_z;
    energy += grid->energy_list[(suf_z+1) + grid->num.z*(suf_y+1) + grid->num.z*grid->num.y*(suf_x+1)] * frac_x     * frac_y     * frac_z;

    return energy;
  }
}
