#include <string>
#include <vector>
#include <fstream>
#include <iostream>
#include <cstdlib>
#include <boost/algorithm/string/trim.hpp>
#include "pdbqt_reader.hpp"
#include "utils.hpp"

namespace {
template<typename T>
T get_field(const std::string &str, int start, int end){
  return utils::cast_substring<T>(str, start-1, end-start+1);
}

template<>
std::string get_field(const std::string &str, int start, int end){
  return str.substr(start-1, end-start+1);
}
} // namespace 

namespace format{
using std::string;

PDBQTMolecule ParsePDBQTFile(std::istream& pdbqt_stream){
  PDBQTMolecule pdbqt_mol;
  pdbqt_mol.title = "";
  while(!pdbqt_stream.eof()) {
    string pdbqt_record;
    std::getline(pdbqt_stream, pdbqt_record);
    string record_name = pdbqt_record.substr(0,6);
    if(record_name == "TITLE ") pdbqt_mol.title += boost::trim_right_copy(pdbqt_record.substr(10, 70));
    if(record_name == "ATOM  " || record_name == "HETATM"){
      format::PDBQTAtom atom;
      atom.id   = get_field<int>(pdbqt_record, 7, 11);
      atom.name = get_field<string>(pdbqt_record,13,16);
      atom.residue_name = get_field<string>(pdbqt_record, 18, 20);
      atom.chainID = get_field<char>(pdbqt_record, 22, 22);
      atom.residueID = get_field<int>(pdbqt_record, 23, 26);
      atom.x = get_field<fltype>(pdbqt_record, 31, 38);
      atom.y = get_field<fltype>(pdbqt_record, 39, 46);
      atom.z = get_field<fltype>(pdbqt_record, 47, 54);
      atom.charge = get_field<fltype>(pdbqt_record, 71, 76);
      atom.element = utils::trim_tail(get_field<string>(pdbqt_record, 78, 79));

      // Hydrogen is OK to contain.
      //if (atom.element == " H"){
      //  std::cerr << "ERROR: unexpected Atom element `H'" << std::endl;
      //  abort();
      //}
      if (atom.residue_name == "HOH") continue;
      pdbqt_mol.atoms.push_back(atom);
    }
    if(record_name == "CONECT"){
      int len = utils::trim_tail(pdbqt_record).size();
      int num = (len-11)/5;
      for(int i=0; i<num; i++){
	format::PDBQTBond bond;
	bond.bond_id = 0;
	bond.origin_atom_id = get_field<int>(pdbqt_record, 7, 11);
	bond.target_atom_id = get_field<int>(pdbqt_record, 12+i*5, 16+i*5);
	if(bond.origin_atom_id > bond.target_atom_id){
	  int temp = bond.target_atom_id;
	  bond.target_atom_id = bond.origin_atom_id;
	  bond.origin_atom_id = temp;
	}
	bond.origin_atom_index = 0;
	bond.target_atom_index = 0;

	bool flag = true;
	for(std::vector<PDBQTBond>::const_iterator it = pdbqt_mol.bonds.begin(); it != pdbqt_mol.bonds.end(); ++it){
	  if(bond.origin_atom_id == it->origin_atom_id && bond.target_atom_id == it->target_atom_id){
	    flag = false;
	    break;
	  }
	}
	if(flag){
	  pdbqt_mol.bonds.push_back(bond);
	}

      }

    }
  }

  

  int count = 0;
  for(std::vector<PDBQTBond>::iterator it = pdbqt_mol.bonds.begin(); it != pdbqt_mol.bonds.end(); ++it){
    it->bond_id = count++;

    int orig_idx;
    for (orig_idx = 0; orig_idx < pdbqt_mol.atoms.size(); orig_idx++) {
      if (it->origin_atom_id == pdbqt_mol.atoms[orig_idx].id) break;
    }
    
    int target_idx;
    for (target_idx = 0; target_idx < pdbqt_mol.atoms.size(); target_idx++) {
      if (it->target_atom_id == pdbqt_mol.atoms[target_idx].id) break;
    }

    if (orig_idx == pdbqt_mol.atoms.size() || target_idx == pdbqt_mol.atoms.size()){
      std::cerr << "ERROR:ParsePDBQTFile: detected inconsistent relation between bonds and atoms." << std::endl;
    } else {
      it->origin_atom_index = orig_idx;
      it->target_atom_index = target_idx;
    }    
  }

  return pdbqt_mol;
}

PDBQTMolecule ParsePDBQTFile(const char *filename){ 
  std::ifstream pdbqt_stream(filename);
  if(pdbqt_stream.fail()){
    std::cerr << "Error: failed to open PDBQT file:" << filename << std::endl;
    abort();
  }
  return format::ParsePDBQTFile(pdbqt_stream);
}
}
