#include <cmath>
#include <vector>
#include <algorithm>
#include <map>
#include <sys/time.h>

#include "pdb_reader.hpp"
#include "mol_reader.hpp"
#include "infile_reader.hpp"
#include "ff_reader.hpp"
#include "paramAD_reader.hpp"
#include "docking.hpp"
#include "utils.hpp"
#include "log_writer.hpp"
#include "convert_format.hpp"
#include "utils.hpp"
#include "energy_grid.hpp"
#include "analysis.hpp"

#include "AtomEnergyGrid.hpp"
#include "EnergyCalculator.hpp"

#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>
#include <fstream>


namespace {

  void MakeRotatedMoleculeList(const fragdock::Molecule& molecule, 
			       const std::vector<docking::RotInfo>& vec_rti,
			       std::vector<fragdock::Molecule>& vec_molecule){
    vec_molecule.clear();
  
    for(std::vector<docking::RotInfo>::const_iterator rti = vec_rti.begin(); rti!=vec_rti.end(); rti++){
      fragdock::Molecule temp = molecule;
      temp.rotate(rti->theta, rti->phi, rti->psi);
      vec_molecule.push_back(temp);
    }
  
  }

  fltype ComputeEnergy(const fragdock::Molecule& receptor, 
		       const fragdock::Molecule& ligand, 
		       const std::vector<fragdock::AtomEnergyGrid>& atom_grids){
    static const fltype ENERGY_LIMIT = 10000;
    fltype sum = 0;
    const std::vector<fragdock::Atom>& atoms = ligand.getAtoms();

    for(int i=0; i<atoms.size(); i++){
      const fragdock::Atom& atom = atoms[i];
      sum += atom_grids[atom.getType().getId()].getEnergy(atom);
      sum += atom_grids[fragdock::AtomType::id_table["e"]].getEnergy(atom) * atom.getCharge();

      if(sum >= ENERGY_LIMIT){
	sum = ENERGY_LIMIT;
	break;
      }
    }
    return sum;
  }
}


namespace docking {

  void MakeTranslations(const fragdock::Point3d<fltype> &center, 
			const fragdock::Point3d<fltype> &pitch, 
			const fragdock::Point3d<int> &num,
			std::vector<fragdock::Vector3d> &translations){
    for(fltype x=0; x<num.x; x++){
      for(fltype y=0; y<num.y; y++){
	for(fltype z=0; z<num.z; z++){
	  fragdock::Vector3d point(center.x+pitch.x*(x-(num.x-1)/2),
				   center.y+pitch.y*(y-(num.y-1)/2),
				   center.z+pitch.z*(z-(num.z-1)/2));
	  translations.push_back(point);
	}
      }
    }  
  }


  void MakeRotations(const std::string &Filename, 
		     std::vector<docking::RotInfo> &rotations){
    //TODO maybe something wrong. reading file is should placed reader.cc or like that.
    std::ifstream ifs(Filename.c_str());
    if(ifs.fail()){
      std::cerr << "opening rotangs file failed:" << Filename << std::endl;
      abort();
    }
    std::string buffer;


    while(!ifs.eof()){
      std::getline(ifs, buffer);
      std::vector<std::string> v;
      boost::algorithm::split(v, buffer, boost::is_any_of(","));
      if(v.size() >= 3){
	fltype x, y, z;
	x = boost::lexical_cast<fltype>(v[0]);
	y = boost::lexical_cast<fltype>(v[1]);
	z = boost::lexical_cast<fltype>(v[2]);
	rotations.push_back((docking::RotInfo){x, y, z});
      }
    }
  }

  int NumNearAtoms(const fragdock::Molecule& receptor, const fragdock::Vector3d& point, fltype threashold_dist){
    int count = 0;
    for(int i=0; i<receptor.size(); i++){
      fltype dist = (receptor.getAtom(i) - point).abs();
      if(dist <= threashold_dist) count++;
    }

    return count;
  }


  int GetNearAtoms(const fragdock::Molecule *receptor, 
		   const fragdock::Vector3d *point,
		   const fltype threashold_dist, 
		   std::vector<fragdock::Atom>* retAtoms){
    int count = 0;
    const bool isRetAtomsNull = (retAtoms == NULL);
    if(!isRetAtomsNull) retAtoms->clear();
    for(std::vector<fragdock::Atom>::const_iterator iter=receptor->getAtoms().begin(); iter!=receptor->getAtoms().end(); iter++){
      if(( *iter - *point ).abs() < threashold_dist){
	if(!isRetAtomsNull) retAtoms->push_back(*iter);
	count++;
      }
    }
    return count;
  }



  //newer version.
  void docking(const fragdock::Molecule& receptor,
	       const fragdock::Molecule& ligand,
	       const std::vector<fragdock::Vector3d>& translations,
	       const std::vector<docking::RotInfo>& rotangs,
	       const std::string& grid_folder,
	       std::vector< std::pair<fltype, Pose> >& conformations,
	       int num_of_outputs){

    int docking_count = 0;
    struct timeval st_time, ed_time;
    gettimeofday(&st_time, NULL);
    const std::vector<fragdock::AtomEnergyGrid> atom_grids = fragdock::AtomEnergyGrid::readAtomGrids(grid_folder);

    /* prepare translations and rotations */
    const int total = translations.size()*rotangs.size(); //The number of docking poses
    const int tri_size = translations.size(); 
    const int rti_size = rotangs.size();

    std::vector<fragdock::Molecule> rotated_molecules;
    MakeRotatedMoleculeList(ligand, rotangs, rotated_molecules);

#pragma omp parallel for shared(docking_count, rotated_molecules)
    for(int i=0; i<tri_size; i++){ //translation loop
      const fragdock::Vector3d& tri = translations.at(i);

      //local_conformations are temporary lists of docked conformations and free energy.
      // *Notice* It doesn't share across omp threads.  
      std::vector< std::pair<fltype, Pose> > local_conformations; 

      for(int j=0; j<rti_size; j++){ //rotation loop
	const docking::RotInfo &rti = rotangs.at(j);
	fragdock::Molecule dockMolecule = rotated_molecules.at(j);

	/* prepare DockMolecule and calculate */
	dockMolecule.translate(tri);
	fltype energy = ComputeEnergy(receptor, dockMolecule, atom_grids);
      

	/* show the docking progress */
#pragma omp atomic
	docking_count++;
      
	if(docking_count%500000 == 0){
	  std::cout << docking_count << "/" << total <<" dockings" << std::endl;
	  gettimeofday(&ed_time, NULL);
	  if(total != docking_count){
	    std::cout << "least time : " 
		      << ((ed_time.tv_sec-st_time.tv_sec)+(ed_time.tv_usec-st_time.tv_usec)*0.001*0.001)/docking_count*(total-docking_count) << " sec." << std::endl;
	  }	
	}


	/* add to local_conformations */
	local_conformations.push_back(std::pair<fltype, Pose>(energy, (Pose){tri,rti}));
	while(local_conformations.size() > num_of_outputs){
	  sort(local_conformations.begin(), local_conformations.end());
	  local_conformations.pop_back();
	}

      } // rotation loop end

      /* all "local_conformations" are aggregated into (global) "conformations". */
#pragma omp critical
      {
	for(std::vector< std::pair<fltype, Pose> >::iterator it=local_conformations.begin();
	    it!=local_conformations.end();
	    ++it){
	  conformations.push_back(*it);
	  while(conformations.size() > num_of_outputs){
	    sort(conformations.begin(), conformations.end());
	    conformations.pop_back();
	  }
	}
      } // critical section end

    } //translation loop end and parallel end.


    sort(conformations.begin(), conformations.end());
    gettimeofday(&ed_time, NULL);

    const fltype dockingTime = ((ed_time.tv_sec-st_time.tv_sec)+(ed_time.tv_usec-st_time.tv_usec)*0.001*0.001);
    logger::writeLog((boost::format("Docking_count: %d dockings")%docking_count).str(), logger::logLevel::INFO);
    std::cout << "Docking_count: " << docking_count << " dockings" << std::endl;
    logger::writeLog((boost::format("Total docking time: %f sec.")%dockingTime).str(), logger::logLevel::INFO);
    std::cout << "Total docking time: " << dockingTime << " sec." << std::endl;
    logger::writeLog((boost::format("Average docking time: %f sec.")%(dockingTime/docking_count)).str(), logger::logLevel::INFO);
    std::cout << "Average docking time: " << (dockingTime/docking_count) << " sec." << std::endl;


    std::cout << "best energy=" << conformations.front().first << std::endl;
    std::cout << "best conformation trinfo={" 
	      << conformations.front().second.trinfo.x << ","
	      << conformations.front().second.trinfo.y << ","
	      << conformations.front().second.trinfo.z << "} rotinfo={"
	      << conformations.front().second.rotinfo.theta << ","
	      << conformations.front().second.rotinfo.phi << ","
	      << conformations.front().second.rotinfo.psi << "}" << std::endl;

    std::cout << "input ligand's energy=" << ComputeEnergy(receptor, ligand, atom_grids) << std::endl;

  }

} // namespace format
